/*________________________________________________________________________________________________

	Title: Convert to RNL Chromaticity
	'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	Authors: Jolyon Troscianko & Cedric van den Berg
	Date: 16/11/2018
............................................................................................................

This script converts from cone-catch quanta to receptor-noise-limited chromaticity space (as described
in Renoult et al 2017; Colour spaces in ecology and evolutionary biology, equations A1.23 to A 1.29.)

Euclidean distances measured in this colour space are identical to JND calculations.

This script uses the log-normal version of the RNL model, up to tetrachromacy.

--------------------------------------------------------------------------------------------------------------------------------
Requirements:

Input images should be cone-catch images with slices labelled according to the receptor channel.
The user is asked to input the Weber fraction specific to each channel.

Any channels with Weber fractions of zero are ignored
_________________________________________________________________________________________________
*/



import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;
import ij.plugin.*;
import java.awt.*;

import ij.plugin.frame.RoiManager;
import ij.gui.Roi; 


public class Extract_RNL_Colour_Map implements PlugInFilter {

ImageStack stack;
	public int setup(String arg, ImagePlus imp) { 
	stack = imp.getStack(); 
	return DOES_32 + STACK_REQUIRED; 
	}



public void run(ImageProcessor ip) {


IJ.showStatus("Convert to RNL");


int w = stack.getWidth();
int h = stack.getHeight();
int nPoints = w*h;
int nDims = stack.getSize();

double[] weber = new double[nDims]; //

//for(int i = 0; i<nDims; i++)
//	weber[i] = 0.05;

String[] sliceNames = new String[nDims];
sliceNames = stack.getSliceLabels();
int nChannels = 0;
int[] chL = new int[nDims]; // channel lookup


	GenericDialog gd = new GenericDialog("Convert to RNL Chromaticity");

		gd.addStringField("Label:", "", 20);
		for(int i=0; i<nDims; i++)
			gd.addNumericField("Weber_" + (i+1) + " " + sliceNames[i], weber[i], 3);	

		gd.addNumericField("Resolution (px per JND): ", 4, 0);	

	gd.showDialog();
	if (gd.wasCanceled())
		return;

String label = gd.getNextString();
for(int i=0; i<nDims; i++){
	weber[i] = gd.getNextNumber();
	if(weber[i] != 0.0){
		IJ.log("" + sliceNames[i] + " " + weber[i]);
		chL[nChannels] = i;
		nChannels++;
	}
}
int res = (int) (gd.getNextNumber());



if(nChannels > 4 || nChannels < 2){
	IJ.log("This tool only supports di, tri, or tetrachromatic images");
	return;
}

float[] pxs = new float[nPoints*nDims];
float[] pixels = new float[nPoints];
int[] zeros = new int[nPoints];


ImagePlus imp = IJ.getImage();
Roi roi = imp.getRoi();

if( roi == null ){ //-----------------No selection - Measuring whole image-----------------
	IJ.log("Measuring whole image");
	for(int k=0; k<w*h; k++)
		zeros[k] = 1;
} else { //-----------------Measuring ROI-----------------
	//IJ.log("Measuring Selection");
	ImageProcessor mask = roi.getMask();

	Rectangle r = roi.getBounds();

	if(mask == null){
		IJ.log("Measuring rectangular selection");
		for (int y=r.y; y<r.y+r.height; y++)
		for (int x=r.x; x<r.x+r.width; x++)
			zeros[(y*w) + x] = 1;
	} else {
		IJ.log("Measuring selected area");
		for (int y=r.y; y<r.y+r.height; y++)
		for (int x=r.x; x<r.x+r.width; x++)
		if(mask.getPixel(x-r.x,y-r.y) != 0)
			zeros[(y*w) + x] = 1;
	}


}

for(int i=0; i<nDims; i++){ // convert float pixel values to double-precision array
	pixels = (float[]) stack.getPixels(i+1);
	for(int k=0; k<nPoints; k++)
		if(pixels[k] > 0)
			pxs[k+nPoints*i] = pixels[k];
		else zeros[k] = 0;
}


//ImageStack outStack = new ImageStack(w, h);

/*
IJ.log("__________________________");

IJ.log("Weber 1: " + weber[chL[0]]);
IJ.log("Weber 2: " + weber[chL[1]]);
IJ.log("Weber 3: " + weber[chL[2]]);
*/


double c0 =0.0;
double c1 = 0.0;
double c2 = 0.0;
double c3 = 0.0;
double A = 0.0;
double a = 0.0;
double b = 0.0;
double c = 0.0;

float xMin = (float) 10E10;
float xMax = (float) -10E10;
float yMin = (float) 10E10;
float yMax = (float) -10E10;
float zMin = (float) 10E10;
float zMax = (float) -10E10;


//-----------X--------------
float[] x = new float[nPoints];
float[] y = new float[nPoints];
float[] z = new float[nPoints];

for(int i=0; i<nPoints; i++)
if(zeros[i] == 1){
	c0 = (double) (pxs[i+nPoints*chL[0]]);
	c1 = (double) (pxs[i+nPoints*chL[1]]);
	x[i] = (float) (   Math.pow(1/(Math.pow(weber[chL[1]],2)+Math.pow(weber[chL[0]], 2)), 0.5)*(Math.log(c0)-Math.log(c1))       );		     

	if(x[i] < xMin) xMin = x[i];
	if(x[i] > xMax) xMax = x[i];
}

//outStack.addSlice("X", x);


//-----------Y--------------
if(nChannels > 2){
//float[] y = new float[nPoints];
for(int i=0; i<nPoints; i++)
if(zeros[i] == 1){
	c0 = (double) (pxs[i+nPoints*chL[0]]);
	c1 = (double) (pxs[i+nPoints*chL[1]]);
	c2 = (double) (pxs[i+nPoints*chL[2]]);
	y[i] = (float) (        Math.pow((Math.pow(weber[chL[1]] , 2)+Math.pow(weber[chL[0]] , 2))/(Math.pow(weber[chL[2]]*weber[chL[1]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2)+Math.pow(weber[chL[1]]*weber[chL[0]],2)) , 0.5)*(Math.log(c2)-(Math.log(c0)*(Math.pow(weber[chL[1]], 2)/(Math.pow(weber[chL[1]], 2)+Math.pow(weber[chL[0]],2)))+Math.log(c1)*(Math.pow(weber[chL[0]],2)/(Math.pow(weber[chL[1]], 2)+Math.pow(weber[chL[0]], 2)))))          );

	if(y[i] < yMin) yMin = y[i];
	if(y[i] > yMax) yMax = y[i];
}

//outStack.addSlice("Y", y);
}

//-----------Z--------------
if(nChannels > 3){
//float[] z = new float[nPoints];
for(int i=0; i<nPoints; i++)
if(zeros[i] == 1){
	c0 = (double) (pxs[i+nPoints*chL[0]]);
	c1 = (double) (pxs[i+nPoints*chL[1]]);
	c2 = (double) (pxs[i+nPoints*chL[2]]);
	c3 = (double) (pxs[i+nPoints*chL[3]]);
	A = Math.pow((Math.pow(weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[1]], 2))/(Math.pow(weber[chL[2]]*weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[3]]*weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[3]]*weber[chL[2]]*weber[chL[0]], 2)+Math.pow(weber[chL[3]]*weber[chL[2]]*weber[chL[1]], 2)),0.5);
	a = Math.pow(weber[chL[2]]*weber[chL[1]], 2)/(Math.pow(weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[1]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2));
	b = Math.pow(weber[chL[2]]*weber[chL[0]], 2)/(Math.pow(weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[1]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2));
	c = Math.pow(weber[chL[1]]*weber[chL[0]], 2)/(Math.pow(weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[1]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2));
	z[i] = (float) (       A*(Math.log(c3)-(a*Math.log(c0)+b*Math.log(c1)+c*Math.log(c2)))     );

	if(z[i] < zMin) zMin = z[i];
	if(z[i] > zMax) zMax = z[i];
}

//outStack.addSlice("Z", z);
}

//new ImagePlus("RNL Chromaticity", outStack).show();



//---------------------Map limits---------------------
int xMapMin = 0;
int xMapMax = 0;
int yMapMin = 0;
int yMapMax = 0;
int zMapMin = 0;
int zMapMax = 0;

// floor or round?

xMapMin = (int) (Math.round(xMin/res)-1)*res;
xMapMax = (int) (Math.round(xMax/res)+1)*res;
if(nChannels > 2){
	yMapMin = (int) (Math.round(yMin/res)-1)*res;
	yMapMax = (int) (Math.round(yMax/res)+1)*res;
}
if(nChannels > 3){
	zMapMin = (int) (Math.round(zMin/res)-1)*res;
	zMapMax = (int) (Math.round(zMax/res)+1)*res;
}


IJ.log("x min: " + xMin + " max: " + xMax);
if(nChannels > 2)
	IJ.log("y min: " + yMin + " max: " + yMax);
if(nChannels > 3)
	IJ.log("z min: " + zMin + " max: " + zMax);

IJ.log("x map min: " + xMapMin + " max: " + xMapMax);
if(nChannels > 2)
	IJ.log("y map min: " + yMapMin + " max: " + yMapMax);
if(nChannels > 3)
	IJ.log("z map min: " + zMapMin + " max: " + zMapMax);


//---------------------------Create Colour Map-----------------------------

int nMap = 0;
if(nChannels == 2)
	nMap = (xMapMax-xMapMin)*res;
if(nChannels == 3)
	nMap = (xMapMax-xMapMin) * (yMapMax-yMapMin) * res * res;
else if(nChannels == 4)
	nMap = (xMapMax-xMapMin) * (yMapMax-yMapMin) * (zMapMax-zMapMin) * res * res * res;
else return;

int[] cMap = new int[nMap];

int nPx = 0;

int xt = 0;
int yt = 0;
int zt = 0;

if(nChannels == 2)
for(int i=0; i<nPoints; i++)
if(zeros[i] == 1){
	cMap[Math.round(x[i]*res)-(xMapMin*res)]++;
	nPx++;
}

if(nChannels == 3)
for(int i=0; i<nPoints; i++)
if(zeros[i] == 1){
	xt = Math.round(x[i]*res)-(xMapMin*res);
	yt = Math.round(y[i]*res)-(yMapMin*res);
	cMap[xt + yt * ((xMapMax-xMapMin)*res) ]++;
	nPx++;
}

if(nChannels == 4)
for(int i=0; i<nPoints; i++)
if(zeros[i] == 1){
	xt = Math.round(x[i]*res)-(xMapMin*res);
	yt = Math.round(y[i]*res)-(yMapMin*res);
	zt = Math.round(z[i]*res)-(zMapMin*res);
	cMap[xt + yt * ((xMapMax-xMapMin)*res) + zt * ((xMapMax-xMapMin)*res) * ((yMapMax-yMapMin)*res)]++;

	//cMap[(Math.round(x[i]*res)-(xMapMin*res))+(Math.round(y[i]*res)-(yMapMin*res)) * (xMapMax-xMapMin)*res + (Math.round(z[i]*res) -(zMapMin*res)) * (xMapMax-xMapMin) *  (yMapMax-yMapMin) * res * res]++;
	nPx++;
}


//--------------------------------Output Colour Map Images----------------------------------------

if(nChannels == 2){
	float[] cMapf = new float[(xMapMax-xMapMin)*res];
	for(int x2=0; x2<(xMapMax-xMapMin)*res; x2++)
		cMapf[x2] = (float) (cMap[x2]);

	ImageStack cMapStack = new ImageStack((xMapMax-xMapMin)*res, 1);
	cMapStack.addSlice("cMap", cMapf);
	new ImagePlus(label +"_Colour_Map", cMapStack).show();
}

if(nChannels == 3){
	float[] cMapf = new float[(xMapMax-xMapMin)*(yMapMax-yMapMin)*res*res];
	for(int y2=0; y2<(yMapMax-yMapMin)*res; y2++)
	for(int x2=0; x2<(xMapMax-xMapMin)*res; x2++)
		cMapf[x2 + y2 * ((xMapMax-xMapMin)*res)] = (float) (cMap[x2 + y2 * ((xMapMax-xMapMin)*res)]);

	ImageStack cMapStack = new ImageStack((xMapMax-xMapMin)*res, (yMapMax-yMapMin)*res);
	cMapStack.addSlice("cMap", cMapf);
	new ImagePlus(label +"_Colour_Map", cMapStack).show();
}
if(nChannels == 4){
	ImageStack cMapStack = new ImageStack((xMapMax-xMapMin)*res, (yMapMax-yMapMin)*res);

	for(int j=0; j<(zMapMax-zMapMin)*res; j++){
		float[] cMapf = new float[(xMapMax-xMapMin)*(yMapMax-yMapMin)*res*res];
		for(int y2=0; y2<(yMapMax-yMapMin)*res; y2++)
		for(int x2=0; x2<(xMapMax-xMapMin)*res; x2++)
			cMapf[x2 + y2 * ((xMapMax-xMapMin)*res)] = (float) (cMap[x2 + y2 * ((xMapMax-xMapMin)*res) + j * ((xMapMax-xMapMin)*res) * ((yMapMax-yMapMin)*res)]);

		cMapStack.addSlice("z" + (zMapMin+j), cMapf);
	}
	new ImagePlus(label +"_Colour_Map", cMapStack).show();
}



ImagePlus imp2 = WindowManager.getCurrentImage();
if (imp2 == null)
	return;

String str = "string=[label=" + label + "";
str = str + ",nPx=" + nPx + "";
str = str + ",res=" + res +"";
str = str + ",channels=" + sliceNames[chL[0]]  + "";
for(int j=1; j<nChannels; j++)
	str = str+ ":" + sliceNames[chL[j]] + "";
str = str + ",weber=" + weber[chL[0]]  + "";
for(int j=1; j<nChannels; j++)
	str = str+ ":" + weber[chL[j]] + "";
str = str + ",x_limits=" + xMapMin + ":" + xMapMax + "";
if(nChannels > 2)
	str = str + ",y_limits=" + yMapMin + ":" + yMapMax + "";
if(nChannels > 3)
	str = str + ",z_limits=" + zMapMin + ":" + zMapMax + "";
str = str+"]";

IJ.run(imp2, "Set Metadata", str);

}
}
