
/*__________________________________________________________________

	Title: Naive Bayes Classifier
	''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	Author: Jolyon Troscianko
	Date: 18/3/2018
............................................................................................................

This code uses a naive Bayes classifier to split an image into a discrete
number of colours.

The classification model uses the mean and standard deviaiton of each
colour type to decide which category to put each colour into.

Values are taken from the results table. The results should match the
output format of measuring ROIs using the "Measure ROIs" script.

i.e. label in the first column, then mean and standard deviation of each
channel repeating in the following columns.

The classifier can handle any number of colour channels. A colour-space
like CIELAB is recommended for standard colour images.

The processed image should be a stack with 32-bits/channel.

____________________________________________________________________
*/


import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.measure.ResultsTable;


public class Naive_Bayes_Classify implements PlugInFilter {

ImageStack stack;
	public int setup(String arg, ImagePlus imp) { 
	stack = imp.getStack(); 
	return DOES_32 + STACK_REQUIRED; 
	}

public void run(ImageProcessor ip) {


	int w = stack.getWidth();
	int h = stack.getHeight();
	int nz = stack.getSize();

	String[] sliceNames = new String[nz];
	sliceNames = stack.getSliceLabels();

	float[] pixels= new float[w*h]; // temp array for loading
	float[] sums= new float[w*h]; // temp array for loading
	float[] tPixels = new float[w*h*nz]; // array containing all pixel values across all slices

	for(int i=0; i<nz; i++){
		pixels = (float[]) stack.getPixels(i+1);
		for(int j=0; j<w*h; j++)
			tPixels[(i*w*h)+j] = pixels[j];
	}

	int[] roiPx = new int[w*h];
	int zeroCount = 0;

	for(int j=0; j<w*h; j++){
		zeroCount = 0;
		for(int m=0; m<nz; m++){
			if(tPixels[(m*w*h)+j] != 0)
				zeroCount++;
		}
		if(zeroCount != 0)
			roiPx[j] = 1;
	}


//-----------------------Get values from Results Table----------------------
// First column should be group IDs, then alternating mean and SD for each dimension

	ResultsTable rt =  ResultsTable.getResultsTable();
	String[] columns = rt.getHeadings();

	double[] tempX = rt.getColumnAsDoubles(1);

	int nResults = tempX.length;
	int nDims = (columns.length-1)/2;

	//IJ.log((columns.length-1)/2 + " ");
	//IJ.log("nDims: " + nDims + " nResults: " + nResults +  " nSlices: " + nz);
	
	double[] means = new double[nResults*nDims];
	double[] sds = new double[nResults*nDims];

//	double temp = rt.getValueAsDouble(0, 0);
//	IJ.log(" " + temp);
//	IJ.log(" " + columns[6]);

	int colOffset = 0; //weird things seem to happen with the first column


	for(int i=0; i<nResults; i++)
	for(int j=0; j<nDims; j++){
		means[(j*nResults)+i] = rt.getValueAsDouble((j*2)+colOffset, i);
		sds[(j*nResults)+i] = rt.getValueAsDouble((j*2)+colOffset+1, i);
	}//ij


	if (nz != nDims){
		IJ.log("Results table and image have different number of dimensions");
		throw new RuntimeException("Results table and image have different number of dimensions");	
	}

//	for(int j=0; j<nDims; j++)
//	for(int i=0; i<nResults; i++)
//		IJ.log(means[(j*nResults)+i] + "_" + sds[(j*nResults)+i]);


//-----------------------Apply Naive Bayes Classifier----------------------



double[] probs = new double[nDims*nResults]; //probabilities for each dimension
double[] pps = new double[nResults]; // partial probabilities for each group

float[] ids = new float[w*h];
double ppTemp = 1.0;
double ppSum = 0.0;
double maxProb = 0.0;
int maxGroup = 0;
int[] idCounts = new int[nResults];

for(int j=0; j<w*h; j++)
if(roiPx[j] == 1){

	for(int l=0; l<nResults; l++)
		for(int m=0; m<nDims; m++){
			//setZCoordinate(m);
			probs[(nResults*m)+l] = Math.pow(((1/(Math.pow(sds[(nResults*m)+l],2)))/(2*Math.PI)),0.5)*Math.exp(         (-1* (1/Math.pow(sds[(nResults*m)+l],2))*Math.pow((tPixels[(m*w*h)+j] - means[(nResults*m)+l]),2))/2   );
		}



	for(int l=0; l<nResults; l++){
		ppTemp = 1;
		ppSum = 0;
		for(int m=0; m<nDims; m++){ //  multiply together the probabilities of all the groups
			ppTemp = ppTemp * probs[(nResults*m)+l];
			ppSum = ppSum + probs[(nResults*m)+l];
		}
		pps[l] = ppTemp;
	}//l
		
	for(int l=0; l<nResults; l++)// calculate overall probabilities for each group
		pps[l] = pps[l]/ppSum;

	maxProb = 0;
	for(int l=0; l<nResults; l++)// Bayes rule - select the most likely group
		if(pps[l] > maxProb){
			maxProb = pps[l];
			maxGroup = l;
		}


	//if(maxProb > 0)
	//	probArray[i] = groupNames[maxGroup];
	ids[j] = maxGroup+1;
	idCounts[maxGroup] ++;

}//j



ImageStack outStack = new ImageStack(w, h);
outStack.addSlice("Group ID", ids);
new ImagePlus("Zone IDs", outStack).show();

IJ.log("Pixel Coverage");
for(int i=0; i<nResults; i++)
	IJ.log(rt.getLabel(i)+ ": \t" + idCounts[i]);


}
}
