/*________________________________________________________________________________________________

	Title: Weber Ranked Filter
	''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	Authors: Jolyon Troscianko & Cedric van den Berg
	Date: 2/10/18
............................................................................................................

Smart-blurs images based on the receptor noise limited model, combining both colour and luminance
with Weber fractions adjustable to any visual system.

Each pixel is compared to all of its neghbours within a given radius. The neighbours are ranked by
chromatic and luminance distances and then exponentially smoothed such that each pixel is combined
most with its nearest neighbours and least with its most distant ones. The result is a colour-based
smart blur that reduces noise while maintaining or enhacing edges.

--------------------------------------------------------------------------------------------------------------------------------
Requirements:

Input images should be 32-bit stack cone-catch images. The image must first contain the chromatic
cone-catch images (colour receptors), followed by the last channel which is the luminance channel.

e.g. human: LWS, MWS, SWS, (average of LWS + MWS)
bluetit: UVS, SWS, MWS, LWS, double cone
dichromat: LWS, SWS, LWS

Each of these makes assumption about what the "luminance" channel is, which may not be particularly
well established, so judgement is required!

_________________________________________________________________________________________________
*/


import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.measure.ResultsTable;
import ij.text.TextWindow;
import java.awt.*;
import ij.gui.*;

//import ij.util.*;

//importClass(Packages.org.apache.commons.math.stat.StatUtils);


public class RNL_Ranked_Filter implements PlugInFilter {

ImageStack stack;
	public int setup(String arg, ImagePlus imp) { 
	stack = imp.getStack(); 
	return DOES_32 + STACK_REQUIRED; 
	}
public void run(ImageProcessor ip) {

IJ.showStatus("JND Smart Blur");



int w = stack.getWidth();
int h = stack.getHeight();
int nPoints = w*h;
int nDims = stack.getSize();


double[] weber = new double[nDims-1]; //

if(nDims == 4){
	weber[0] = 0.05; // LW HUMAN scaled to 0.05 for LWS from Hofer et al 2005, assuming a 2:1 L:M and 5.72% S cones
	weber[1] = 0.0707106781; // MW
	weber[2] = 0.1657433633; // SW
} else {
	for(int i = 0; i<nDims-1; i++)
		weber[i] = 0.05;
}

String[] sliceNames = new String[nDims];
sliceNames = stack.getSliceLabels();

double lumWeber = 0.05;

int nRepeats = 1;
int radius = 1;
double falloff = 2;

	GenericDialog gd = new GenericDialog("Weber Ranked Filter");
			gd.addMessage("Weber Fractions:");
			for(int i=0; i<nDims-1; i++)
				gd.addNumericField("Weber_" + (i+1) + " " + sliceNames[i], weber[i], 3);	

			gd.addNumericField("Lum", lumWeber, 4);
			gd.addMessage("");
			gd.addNumericField("Iterations", nRepeats, 0);
			gd.addMessage("A larger radius means a larger number of\nneighbouring pixels are used to influence\nthe smoothing");
			gd.addNumericField("Radius", radius, 0);
			gd.addMessage("Higher falloff values create a steeper falloff\nin the influence of neighbouring pixels,\npreserving more fine detail");
			gd.addNumericField("Falloff", falloff, 4);

	gd.showDialog();
	if (gd.wasCanceled())
		return;

	for(int i=0; i<nDims-1; i++)
		weber[i] = gd.getNextNumber();
	
	lumWeber = gd.getNextNumber();
	nRepeats = (int) gd.getNextNumber();
	radius = (int) gd.getNextNumber();
	falloff = gd.getNextNumber();

int windowSize = (radius * 2) + 1; // e.g. a radius of 1 should be 9 squares
windowSize = windowSize*windowSize; // convert to the square window dimensions required

double[] dfis = new double[nDims-1];
double JND = 0; 

double[] sample1 = new double[nDims-1];
double[] sample2 = new double[nDims-1];

double lum1 = 0.0;
double lum2 = 0.0;


//------------Import Pixel values---------------

float[] tPxs = new float[w*h]; // temp array for loading
double[] pxs = new double[nPoints*nDims];
double[] smPxs = new double[nPoints*nDims];


for(int i=0; i<nDims; i++){ // convert float pixel values to double-precision array
	tPxs = (float[]) stack.getPixels(i+1);
	for(int k=0; k<nPoints; k++)
		if(tPxs[k] > 0) // don't import negative numbers
			pxs[(i*nPoints)+k] = (double) tPxs[k];
		else pxs[k] = 0; // flag negative numbers in the first channel

}


// CALCULATE NEAREST COLOUR

double lowestJND = 0.0;
int lowestLoc = 0;

double[] jnds = new double[windowSize];
double[] vals = new double[windowSize*nDims];

int[] jndRank = new int[windowSize];
double[] weights = new double[windowSize];


// CREATE SIMILARITY WEIGHTING CURVE
double weightsSum = 0.0;
//double tempVal = (double) (1.0/windowSize);
for(int i=0; i<windowSize; i++){
	weights[windowSize-i-1] = Math.pow((double) (i * (1.0/windowSize)), falloff);
	weightsSum = weightsSum + weights[windowSize-i-1];
}

for(int i=0; i<windowSize; i++){
	weights[i] = weights[i]/weightsSum;
	//IJ.log(" " + (float) weights[i]);
}

double[] tmps = new double[nDims];

int windowLoc = 0;
//int windowCount = 0;

double colJND = 0.0;
double lumJND = 0.0;


for(int i=0; i<nRepeats; i++){

IJ.showStatus("Weber Ranked Filter " + (i+1) +" of " + nRepeats);
for(int y=0; y<h; y++){
	IJ.showProgress(((double) ((y+1.0)/h)));
	for(int x=0; x<w; x++)
	if(pxs[(y*w)+x] > 0.0){  //ignore zeros
		windowLoc = 0;

		for(int y2 = y-radius; y2<= y+radius; y2++)
		if(y2 >= 0 && y2 < h)
			for(int x2 = x-radius; x2<= x+radius; x2++)
			if(x2 >= 0 && x2 < w)
			if(pxs[(y2*w)+x2] > 0.0){  //ignore zeros

				for(int k=0; k<nDims-1; k++)
					dfis[k] = Math.log(pxs[(k*nPoints)+(y*w)+x]/pxs[(k*nPoints)+(y2*w)+x2]);

				if(nDims == 3) // DICHROMATIC EQUATION
					colJND = Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5);

				else if(nDims == 4) // TRICHROMATIC EQUATION
					colJND = Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5);

				else if(nDims == 5) // TETRACHROMATIC EQUATION
					colJND= Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5);

				else {
					IJ.log("Image must have 2, 3 or 4 colour receptor channels plus luminance in the last channel");
					return;
				}

				lum2 = pxs[((nDims-1)*nPoints)+(y2*w)+x2];
				lumJND=Math.abs((Math.log(pxs[((nDims-1)*nPoints)+(y*w)+x])-Math.log(lum2))/lumWeber);
			

				// ---------------combine luminance and colour JNDs into a single distance--------------------
				jnds[windowLoc] = Math.pow( Math.pow(colJND, 2) + Math.pow(lumJND, 2) , 0.5);


				for(int k=0; k<nDims; k++)
					vals[(k*windowSize)+windowLoc] = pxs[(k*nPoints)+(y2*w)+x2];

				windowLoc ++;


			}

		// RANK JNDS

			for(int j=0; j<windowLoc; j++)
				jndRank[j] = -1;

			for(int k=0; k<windowLoc; k++){
				lowestJND = 10E10;
				for(int j=0; j<windowLoc; j++)
					if(jnds[j] < lowestJND && jndRank[j] == -1){
						lowestJND = jnds[j];
						lowestLoc = j;
					}//fi
				
				jndRank[lowestLoc] = k;
				for(int l=0; l<nDims; l++)
					tmps[l] = tmps[l] + (vals[(l*windowSize)+lowestLoc]*weights[k]);
				
			}//k
			
			for(int k=0; k<nDims; k++){
				smPxs[(nPoints*k)+(y*w)+x] = tmps[k];
				tmps[k] = 0.0;
			}

	}//x
}//y

	for(int k=0; k<nDims; k++)
		for(int j=0; j<nPoints; j++)
			pxs[(k*nPoints) + j] = smPxs[(k*nPoints) + j];


}//i


ImageStack outStack = new ImageStack(w, h);

for(int i=0; i<nDims; i++){
	float[] fPixels= new float[w*h]; // re-initialise array
	for(int k=0; k<nPoints; k++)
		fPixels[k] = (float) (pxs[(i*nPoints)+k]);
	outStack.addSlice(sliceNames[i], fPixels);
}//i

new ImagePlus("Output", outStack).show();


}
}
